use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::collections::HashSet;


const DAY: &str = "day16";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
enum Tile {
    Empty, ForwardMirror, BackwardMirror, HorizontalSplitter, VerticalSplitter
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
enum Direction {
    North, East, South, West
}

impl Tile {
    fn apply(&self, direction: &Direction) -> (Direction, Option<Direction>) {
        match (self, direction) {
            (Tile::Empty, _) => (*direction, None),
            
            (Tile::ForwardMirror, Direction::North) => (Direction::East, None),
            (Tile::ForwardMirror, Direction::East) => (Direction::North, None),
            (Tile::ForwardMirror, Direction::South) => (Direction::West, None),
            (Tile::ForwardMirror, Direction::West) => (Direction::South, None),
            
            (Tile::BackwardMirror, Direction::North) => (Direction::West, None),
            (Tile::BackwardMirror, Direction::West) => (Direction::North, None),
            (Tile::BackwardMirror, Direction::South) => (Direction::East, None),
            (Tile::BackwardMirror, Direction::East) => (Direction::South, None),
            
            (Tile::HorizontalSplitter, Direction::North) => (Direction::West, Some(Direction::East)),
            (Tile::HorizontalSplitter, Direction::West) => (Direction::West, None),
            (Tile::HorizontalSplitter, Direction::South) => (Direction::East, Some(Direction::West)),
            (Tile::HorizontalSplitter, Direction::East) => (Direction::East, None),
            
            (Tile::VerticalSplitter, Direction::North) => (Direction::North, None),
            (Tile::VerticalSplitter, Direction::West) => (Direction::North, Some(Direction::South)),
            (Tile::VerticalSplitter, Direction::South) => (Direction::South, None),
            (Tile::VerticalSplitter, Direction::East) => (Direction::South, Some(Direction::North)),
        }
    }
    
    fn try_from_byte(byte: u8) -> Option<Self> {
        match byte {
            b'.' => Some(Self::Empty),
            b'/' => Some(Self::ForwardMirror),
            b'\\' => Some(Self::BackwardMirror),
            b'-' => Some(Self::HorizontalSplitter),
            b'|' => Some(Self::VerticalSplitter),
            _ => None,
        }
    }
}

impl Direction {
    fn apply(&self, x: usize, y: usize, max_x: usize, max_y: usize) -> Option<(usize, usize)> {
        match self {
            Self::North => y.checked_sub(1).map(|y| (x,y)),
            Self::West => x.checked_sub(1).map(|x| (x,y)),
            Self::South => (y < max_y).then_some((x,y + 1)),
            Self::East => (x < max_x).then_some((x + 1, y)),
        }
    }
}

fn reflection_energized(map: &Vec<Vec<Tile>>, start: (usize, usize, Direction)) -> usize {
    let (max_x, max_y) = (map[0].len() - 1, map.len() - 1);
    
    let mut visited = HashSet::from([start]);
    let mut frontier = vec![start];
    
    while let Some((x,y,dir)) = frontier.pop() {
        let (dir, extra_dir) = map[y][x].apply(&dir);
        if let Some((x,y)) = dir.apply(x,y, max_x, max_y) {
            if visited.insert((x, y, dir)) {
                frontier.push((x, y, dir));
            }
        }
        if let Some(dir) = extra_dir {
            if let Some((x,y)) = dir.apply(x,y, max_x, max_y) {
                if visited.insert((x,y,dir)) {
                    frontier.push((x, y, dir));
                }
            }
        }
    }
    
    let visited: HashSet<_> = HashSet::from_iter(visited.into_iter().map(|(x,y,_)| (x,y)));
    return visited.len();
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let map: Vec<Vec<Tile>> = input
        .into_iter()
        .map(|line| line
            .as_bytes()
            .into_iter()
            .map(|byte| Tile::try_from_byte(*byte).unwrap())
            .collect::<Vec<_>>())
        .collect::<Vec<_>>();
    
    let solution1 = reflection_energized(&map, (0, 0, Direction::East));
    
    let (max_x, max_y) = (map[0].len() - 1, map.len() - 1);
    
    
    let mut solution2 = solution1;
    for x in 0..=max_x {
        solution2 = solution2.max(reflection_energized(&map, (x, 0, Direction::South)));
        solution2 = solution2.max(reflection_energized(&map, (x, max_y, Direction::North)));
    }
    for y in 0..=max_y {
        solution2 = solution2.max(reflection_energized(&map, (0, y, Direction::East)));
        solution2 = solution2.max(reflection_energized(&map, (max_x, y, Direction::West)));
    }
    

    Ok([Some(solution1.to_string()), Some(solution2.to_string())])
}

#[test]
pub fn test_day16() {
    assert!(common::run_test(DAY, &run))
}
