use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::collections::{BTreeMap, BinaryHeap};
use std::cmp::Reverse;
use std::io::Write;

const DAY: &str = "day17";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

#[derive(PartialEq, Eq, Copy, Clone, Debug, PartialOrd, Ord)]
enum Direction {
    North, East, South, West, NoDirection
}

impl Direction {
    fn delta(self: &Self, x: usize, y: usize, end_x: usize, end_y: usize) -> Option<(usize, usize)> {
        match self {
            Direction::North => {
                if y > 0 {
                    Some((x, y - 1))
                } else {
                    None
                }
            }
            Direction::East => {
                if x < end_x {
                    Some((x + 1, y))
                } else {
                    None
                }
            },
            Direction::South => {
                if y < end_y {
                    Some((x, y + 1))
                } else {
                    None
                }
            },
            Direction::West => {
                if x > 0 {
                    Some((x - 1, y))
                } else {
                    None
                }
            },
            Direction::NoDirection => None
        }
    }
    
    fn left(&self) -> Self {
        match self {
            Direction::North => Direction::West,
            Direction::West => Direction::South,
            Direction::South => Direction::East,
            Direction::East => Direction::North,
            Direction::NoDirection => Direction::East,
        }
    }
    
    fn right(&self) -> Self {
        match self {
            Direction::West => Direction::North,
            Direction::South => Direction::West,
            Direction::East => Direction::South,
            Direction::North => Direction::East,
            Direction::NoDirection => Direction::South,
        }
    }
}

fn do_move(prev_dir: Direction, new_dir: Direction, straight_length: usize) -> Option<(Direction, usize)> {
    let mut log = std::io::stdout();
    if prev_dir == new_dir {        
        if straight_length >= 3 {
            // write!(log, " l>=3 SKIP]");
            return None
        }
        Some((new_dir, straight_length + 1))
    } else {
        Some((new_dir, 1))
    }
}

fn do_ultra_move(prev_dir: Direction, new_dir: Direction, straight_length: usize) -> Option<(Direction, usize)> {
    let mut log = std::io::stdout();
    if prev_dir == new_dir {        
        if straight_length >= 10 {
            // write!(log, " l>=10 SKIP]");
            return None
        }
        Some((new_dir, straight_length + 1))
    } else if prev_dir != Direction::NoDirection && straight_length < 4 {
        // write!(log, " l<4 SKIP]");
        None
    } else {
        Some((new_dir, 1))
    }
}

fn search(map: &Vec<Vec<usize>>, end_x: usize, end_y: usize, 
            h: &dyn Fn(usize, usize) -> usize, 
            do_move: &dyn Fn(Direction, Direction, usize) -> Option<(Direction, usize)>) 
    -> Option<usize> 
{
    let mut frontier = BinaryHeap::new();
    frontier.push(Reverse((h(0,0) + 0, 0, (0,0,Direction::NoDirection, 0), (0,0,Direction::NoDirection, 0))));

    let mut scores = BTreeMap::new();
    scores.insert((0,0,Direction::NoDirection, 0), 0);
    
    let mut parents = BTreeMap::new();
    
    let mut final_cost = None;
    
    let mut log = std::io::stdout();
    
    while let Some(Reverse((h_score, cost, (x, y, dir, l), parent))) = frontier.pop() {
        if x == end_x && y == end_y {
            // write!(log, "\nReached {:?} at cost {}\n", (end_x, end_y), cost);
            final_cost = Some(cost);
            let mut node = &(x,y,dir,l);
            while let Some(parent) = parents.get(node) {
                // write!(log, "{:?}<-", parent);
                node = parent;
            }
            break;
        }
        
        if let Some(g_score) = scores.get(&(x, y, dir, l)) {
            if *g_score < cost {
                continue;
            } else if *g_score == cost && parents.contains_key(&(x, y, dir, l)) {
                // write!(log, "\n{:?}: {}+{}", (x, y, dir, l), cost, h_score);
                break;
            }
        }
        
        parents.insert((x, y, dir, l), parent);
        // write!(log, "\n{:?}: {}+{}", (x, y, dir, l), cost, h_score);
        
        let parent = (x, y, dir, l);
        for direction in [dir.left(), dir.right(), dir].iter() {
            // write!(log, " [ ->{:?}", direction);
            if let Some((direction, l)) = do_move(dir, *direction, l) {
                if let Some((x, y)) = direction.delta(x, y, end_x, end_y) {
                    let candidate_g_score = cost + map[y][x];
                    // write!(log, "={:?}", (x,y));
                    if let Some(g_score) = scores.get(&(x, y, direction, l)) {
                        if candidate_g_score >= *g_score {
                            // write!(log, " had lower path({}<{}) SKIP]", g_score, candidate_g_score);
                            continue;
                        }
                    }
                    scores.insert((x, y, direction, l), candidate_g_score);
                    let f_score = candidate_g_score + h(x,y);
                    frontier.push(Reverse((f_score, candidate_g_score, (x, y, direction, l), parent)));
                    // write!(log, "{:?} INSERTED]", (f_score, candidate_g_score));
                } else {
                    // write!(log, " {:?} at bound SKIP]", (x,y));
                }
            }
        }
//         log.flush();
    }
    return final_cost;
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let map = input.into_iter().map(|s| s.chars().map(|c| c.to_digit(10).unwrap() as usize).collect::<Vec<_>>()).collect::<Vec<Vec<_>>>();
    
    let (end_x, end_y) = (map[0].len() - 1, map.len() - 1);

    assert!(end_x == end_y);
    let midpoint = (end_x + 1)/2;
    
    let radial_h_scores :Vec<usize> = (0..=midpoint)
        .into_iter()
        .map(|i| map[i..=end_x-i].into_iter().fold(usize::MAX, |acc, row| (*row[i..=end_x-i].into_iter().min().unwrap()).min(acc)))
        .fold((vec![], 0), |(mut vec, acc), val| {
            vec.push(acc);
            (vec, acc + val)
        }).0;
    
    println!("{:?}", radial_h_scores);
    
    let h = |x, y| {
        let x_to_edge = midpoint - (midpoint.abs_diff(x));
        let y_to_edge = midpoint - (midpoint.abs_diff(y));
        if x_to_edge <= y_to_edge {
            let cost_to_edge = radial_h_scores[x_to_edge];
            let cost_over_mid = if x <= midpoint { 2 * (midpoint - x) } else { 0 };
            cost_to_edge + (end_y - y) + cost_over_mid
        } else {
            let cost_to_edge = radial_h_scores[y_to_edge];
            let cost_over_mid = if y <= midpoint { 2 * midpoint.abs_diff(y) } else { 0 };
            cost_to_edge + (end_x - x) + cost_over_mid
        }
    };

    let cost = search(&map, end_x, end_y, &h, &do_move);
    let ultra_cost = search(&map, end_x, end_y, &h, &do_ultra_move);
    
    if cost.is_none() {
        return Err("No path found".into());
    }
    
    if ultra_cost.is_none() {
        return Err("No ultra path found".into());
    }
    
    Ok([Some(cost.unwrap().to_string()), Some(ultra_cost.unwrap().to_string())])
}

#[test]
pub fn test_day17() {
    assert!(common::run_test(DAY, &run))
}
