use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::collections::HashSet;

const DAY: &str = "day18";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
enum Curvature {
    Right = -1, Straight = 0, Left = 1
}

impl Curvature {
    fn as_number(&self) -> isize {
        *self as isize
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
enum Direction {
    North, East, South, West
}

const DIRECTIONS: [Direction;4] = [Direction::North, Direction::East, Direction::South, Direction::West];

impl Direction {    
    fn curvature(&self, other: &Direction) -> Curvature {
        if *self == *other {
            Curvature::Straight
        } else if self.left() == *other {
            Curvature::Left
        } else {
            Curvature::Right
        }
    }
    
    fn do_move(&self, (x, y): (isize, isize)) -> (isize, isize) {
        match self {
            Self::North => (x, y - 1),
            Self::East => (x + 1, y),
            Self::South => (x, y + 1),
            Self::West => (x - 1, y),
        }
    }
    
    fn left(&self) -> Self {
        match self {
            Self::North => Self::West,
            Self::East => Self::North,
            Self::South => Self::East,
            Self::West => Self::South
        }
    }
    
    fn right(&self) -> Self {
        match self {
            Self::West => Self::North,
            Self::North => Self::East,
            Self::East => Self::South,
            Self::South => Self::West
        }
    }
    
    fn from_str(s: &str) -> Option<Self> {
        match s {
            "U" => Some(Self::North),
            "L" => Some(Self::West),
            "D" => Some(Self::South),
            "R" => Some(Self::East),
            _ => None
        }
    }
}



pub fn run(input: &Vec<String>) -> AOCResult {
    let mut curvature: isize = 0;
    let mut prev_direction: Option<Direction> = None;
    let (mut x, mut y) = (0, 0);
    
    let mut left_outline = HashSet::new();
    let mut line = HashSet::new();
    let mut right_outline = HashSet::new();
    
    
    for instruction in input.into_iter() {
        let (direction, rest) = instruction.split_once(' ').unwrap();
        let (length, color) = rest.split_once(' ').unwrap();
        let direction = Direction::from_str(direction).unwrap();
        let length: isize = length.parse()?;
        
        let turn_curvature = prev_direction.map(|d| d.curvature(&direction)).unwrap_or(Curvature::Straight);
        prev_direction = Some(direction);
        
        match turn_curvature {
            Curvature::Left => {
                right_outline.insert(direction.right().do_move((x,y)));
            }
            Curvature::Right => {
                left_outline.insert(direction.right().do_move((x,y)));
            }
            _ => {}
        }
        
        for _ in 0..length {
            (x, y) = direction.do_move((x, y));
            line.insert((x,y));
            right_outline.insert(direction.right().do_move((x, y)));
            left_outline.insert(direction.left().do_move((x, y)));
        }
        
        curvature += turn_curvature.as_number();
    }
    
    let outline = if curvature.signum() == (Curvature::Left as isize) {
        left_outline
    } else {
        right_outline
    };
    
    let mut visited = HashSet::new();
    let mut inner_count = line.len();
    
    let mut frontier = vec![];
    for coord in outline.into_iter() {
        if visited.contains(&coord) || line.contains(&coord) {
            continue
        }
        frontier.push(coord);
        while let Some(coord) = frontier.pop() {
            if visited.contains(&coord) || line.contains(&coord) {
                continue
            }
            if !line.contains(&coord) {
                visited.insert(coord);
                inner_count += 1;
            }
            for direction in DIRECTIONS.iter() {
                let coord = direction.do_move(coord);
                if !visited.contains(&coord) && !line.contains(&coord) {
                    frontier.push(coord);
                }
            }
        }
    }
    
    let mut box_stack = vec![];
    let mut curvature: isize = 0;
    let mut start_direction = None;
    let mut prev_direction = None;
    let mut box_points = [None; 4]
    for instruction in input.into_iter() {
        let (direction, rest) = instruction.split_once(' ').unwrap();
        let (length, color) = rest.split_once(' ').unwrap();
        let direction = Direction::from_str(direction).unwrap();
        let length: isize = length.parse()?;
        
        let turn_curvature = prev_direction.map(|d| d.curvature(&direction)).unwrap_or(Curvature::Straight);
        prev_direction = Some(direction);
        start_direction = start_direction.or(prev_direction);
        
        curvature += turn_curvature.as_number();
        
        box_points[curvature.abs()]
        
    }

    Ok([Some(inner_count.to_string()), None])
}

#[test]
pub fn test_day18() {
    assert!(common::run_test(DAY, &run))
}
