use aoc2023_niels_overkamp::common::{self, AOCResult};
use nom::sequence::{preceded, terminated, tuple};
use nom::bytes::complete::tag;
use nom::character::complete::{u32 as u32_parser, alpha1};
use nom::error::convert_error;
use nom::multi::separated_list0;
use nom::combinator::{all_consuming, eof};
use nom::Finish;
use std::mem;

const DAY: &str = "day5";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut seeds_parser = all_consuming(preceded(tag("seeds: "), separated_list0(tag(" "), u32_parser)));
    let mut seeds_parser = |line| seeds_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));
        
    let mut map_parser = all_consuming(tuple((terminated(alpha1, tag("-to-")), terminated(alpha1, tag(" map:")))));
    let mut map_parser = |line| map_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));
        
    let mut range_parser = all_consuming(tuple((terminated(u32_parser, tag(" ")), terminated(u32_parser, tag(" ")), u32_parser)));
    let mut range_parser = |line| range_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));
        
    let empty_parser = eof;
    let empty_parser = |line| empty_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));
    
    let mut input_iter = input.into_iter();
    let mut line = input_iter.next();
    
    let (_, seed_list) = seeds_parser(line.expect("First line").as_str())?;
    
    let mut new_seeds = seed_list.iter().map(|s| Some(*s)).collect::<Vec<_>>();
    let mut seeds = Vec::with_capacity(new_seeds.len());
    seeds.resize(new_seeds.len(), None);
    
    let mut new_seed_ranges = seed_list.chunks_exact(2)
        .map(|a| [a[0], a[1]])
        .map(|[start_x, length]| Some(start_x..(start_x).saturating_add(length))).collect::<Vec<_>>();
    let mut seed_ranges = Vec::with_capacity(new_seed_ranges.len());
    seed_ranges.resize(new_seed_ranges.len(), None);
    
    let apply_map = |from, to, seed: u32| if to > from {
            seed.saturating_add(to - from)
        } else {
            seed.saturating_sub(from - to)
        };
        
    loop {
        line = input_iter.next();
        if line == None {
            break;
        }
        let line = line.unwrap().as_str();
        if let Ok(_) = empty_parser(line) {
            continue;
        } else if let Ok((_,(from, to))) = map_parser(line) {
            for (i, seed_cell) in seeds.iter_mut().enumerate() {
                if let Some(_) = seed_cell {
                    mem::swap(&mut new_seeds[i], seed_cell);
                }
            }
            mem::swap(&mut seeds, &mut new_seeds);
            
            for (i, seed_range) in seed_ranges.iter_mut().enumerate() {
                if let Some(_) = seed_range {
                    mem::swap(&mut new_seed_ranges[i], seed_range);
                }
            }
            mem::swap(&mut seed_ranges, &mut new_seed_ranges);
//             println!("{:?}", seed_ranges);
//             println!("{} -> {}", from, to);
        } else if let Ok((_, (to, from, count))) = range_parser(line) {
            let map_range = from..from.saturating_add(count);
//             println!("{:?} -> {:?}", map_range, to..to.saturating_add(count));
            for (i, seed_cell) in seeds.iter_mut().enumerate() {
                if let Some(seed) = seed_cell {
                    if map_range.contains(seed) {
                        let new_seed = apply_map(from, to, *seed);
                        let _  = seed_cell.take();
                        new_seeds[i] = Some(new_seed);
                    }
                }
            }
            
            let mut i = 0;
            while i < seed_ranges.len() {
                let seed_range = &mut seed_ranges[i];
                if let Some(range) = seed_range {
                    let intersection = range.start.max(map_range.start)..range.end.min(map_range.end);
                    
                    if !intersection.is_empty() {
                        let preamble = range.start..map_range.start.min(range.end);
                        let postamble = map_range.end.max(range.start)..range.end;
                        
                        range.start = apply_map(from, to, intersection.start);
                        range.end = apply_map(from, to, intersection.end);
                        mem::swap(&mut new_seed_ranges[i], seed_range);
                        
                        if !preamble.is_empty() {
                            seed_ranges.push(Some(preamble));
                            new_seed_ranges.push(None);
                        }
                        if !postamble.is_empty() {
                            seed_ranges.push(Some(postamble));
                            new_seed_ranges.push(None);
                        }
                        
                    }
                }
                i += 1;
            }
        }
    }
    
    for (i, seed_cell) in seeds.iter_mut().enumerate() {
        if let Some(_) = seed_cell {
            mem::swap(&mut new_seeds[i], seed_cell);
        }
    }
    
    for (i, seed_range) in seed_ranges.iter_mut().enumerate() {
        if let Some(_) = seed_range {
            mem::swap(&mut new_seed_ranges[i], seed_range);
        }
    }

    let lowest_location = new_seeds.into_iter().map(|seed| seed.unwrap()).min().unwrap();
    let lowest_range_location = new_seed_ranges.into_iter().map(|seed_range| seed_range.unwrap().start).min().unwrap();
    
    Ok([Some(lowest_location.to_string()), Some(lowest_range_location.to_string())])
}

#[test]
pub fn test_day5() {
    assert!(common::run_test(DAY, &run))
}
