use aoc2023_niels_overkamp::common::{self, AOCResult};
use std::collections::HashMap;

const DAY: &str = "day7";

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

const EMPTY_COUNTS: [u8; 14] = [0; 14];

fn to_number(c: char) -> u8 {
    if let Some(n) = c.to_digit(10) {
        n as u8
    } else {
        match c {
            'A' => 14,
            'K' => 13,
            'Q' => 12,
            'J' => 11,
            'T' => 10,
            _ => panic!("Unknown card {}", c)
        }
    }
}


fn to_number_joker(c: char) -> u8 {
    if let Some(n) = c.to_digit(10) {
        n as u8
    } else {
        match c {
            'A' => 14,
            'K' => 13,
            'Q' => 12,
            'J' =>  1,
            'T' => 10,
            _ => panic!("Unknown card {}", c)
        }
    }
}

fn encode(hand: Vec<u8>) -> (Type, [u8; 5]) {
    if let [c1, c2, c3, c4, c5] = &hand[..] {
        (Type::determine(&hand), [*c1, *c2, *c3, *c4, *c5])
    } else {
        panic!("Hand {:?} does not have 5 cards", hand);
    }
}

#[derive(PartialEq, Eq, PartialOrd, Ord, Debug)]
enum Type {
    FiveOfAKind = 7, FourOfAKind = 6, FullHouse = 5, ThreeOfAKind = 4, TwoPair = 3, OnePair = 2, HighCard = 1
}

impl Type {
    fn determine(hand: &[u8]) -> Type {
        let mut counts = Vec::from(EMPTY_COUNTS);
        for card in hand.iter() {
            counts[*card as usize - 1] += 1;
        }
        
        let [max1, max2] = counts[1..].iter().fold([0, 0], |[max1, max2], count| if max1 < *count {
            [*count, max1]
        } else if max2 < *count {
            [max1, *count]
        } else {
            [max1, max2]
        });
        
        let joker_count = counts[0];
        
        match [max1, max2, joker_count] {
            [5, 0, 0] => Type::FiveOfAKind,
            [4, 0, 1] => Type::FiveOfAKind,
            [3, 0, 2] => Type::FiveOfAKind,
            [2, 0, 3] => Type::FiveOfAKind,
            [1, 0, 4] => Type::FiveOfAKind,
            [0, 0, 5] => Type::FiveOfAKind,
            [4, 1, 0] => Type::FourOfAKind,
            [3, 1, 1] => Type::FourOfAKind,
            [2, 1, 2] => Type::FourOfAKind,
            [1, 1, 3] => Type::FourOfAKind,
            [3, 2, 0] => Type::FullHouse,
            [2, 2, 1] => Type::FullHouse,
            [3, 1, 0] => Type::ThreeOfAKind,
            [2, 1, 1] => Type::ThreeOfAKind,
            [1, 1, 2] => Type::ThreeOfAKind,
            [2, 2, 0] => Type::TwoPair,
            [2, 1, 0] => Type::OnePair,
            [1, 1, 1] => Type::OnePair,
            [1, 1, 0] => Type::HighCard,
            _ => panic!("Invalid hand counts: {:?}", [max1, max2])
        }
    }
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut hands = Vec::new();
    let mut joker_hands = Vec::new();
    for line in input.into_iter() {
        let (hand_str, bid) = line.split_once(' ').unwrap();
        let bid = bid.parse::<u32>().unwrap();
        
        let hand = encode(hand_str.chars().map(|c| to_number(c)).collect());
        let joker_hand = encode(hand_str.chars().map(|c| to_number_joker(c)).collect());
        
        hands.push((hand, bid));
        joker_hands.push((joker_hand, bid));
    }
    hands.sort();
    joker_hands.sort();
    
    let winnings = hands.into_iter().enumerate().fold(0, |sum, (i, (_, bid))| sum + (i as u32 + 1) * bid);
    let joker_winnings = joker_hands.into_iter().enumerate().fold(0, |sum, (i, (_, bid))| sum + (i as u32 + 1) * bid);

    Ok([Some(winnings.to_string()), Some(joker_winnings.to_string())])
}

#[test]
pub fn test_day7() {
    assert!(common::run_test(DAY, &run))
}
