use aoc2023_niels_overkamp::common::{self, AOCResult};
use nom::combinator::all_consuming;
use nom::sequence::{terminated, tuple};
use nom::bytes::complete::tag;
use nom::character::complete::alpha1;
use nom::error::convert_error;
use nom::Finish;
use std::collections::{HashSet, HashMap};

const DAY: &str = "day8";
const BYTE_OFFSET: u8 = b'A';

fn main() -> Result<(), Box<dyn std::error::Error>> {
    common::run(DAY, &run)
}

fn to_number(s: &str) -> u16 {
    if let [b1, b2, b3] = s.as_bytes() {
//         println!("{} {} {}", b1, b2, b3);
        ((*b1 - BYTE_OFFSET) as u16) * 26 * 26 + ((*b2 - BYTE_OFFSET) as u16) * 26 + ((*b3 - BYTE_OFFSET) as u16) 
    } else {
        panic!("Identifier is not 3 chars: {}", s)
    }
}

fn bezout_identity(a: i128, b: i128) -> (i128, i128, i128) {
    let mut r0 = a;
    let mut r1 = b;
    let mut s0 = 1;
    let mut t0 = 0;
    let mut s1 = 0;
    let mut t1 = 1;
    
    while r1 != 0 {
        let q = r0 / r1;
        (r0, r1) = (r1, r0 - q * r1);
        (s0, s1) = (s1, s0 - q * s1);
        (t0, t1) = (t1, t0 - q * t1);
    }
    
    (s0, t0, r0)
}

pub fn run(input: &Vec<String>) -> AOCResult {
    let mut node_parser = all_consuming(tuple((terminated(alpha1, tag(" = (")), terminated(alpha1, tag(", ")), terminated(alpha1, tag(")")))));
    let mut node_parser = |line| node_parser(line)
        .finish()
        .map_err(|e| convert_error(line, e));

    let mut graph = Vec::with_capacity(26usize.pow(3));
    graph.resize(26usize.pow(3), [0, 0]);
    let directions = input[0].clone();
    
    let end = to_number("ZZZ") as usize;
    let mut ghost_starts = Vec::new();
    let mut ghost_ends = HashSet::new();
    let mut has_human_end = false;
    
    for node in input[2..].iter() {
        let (_, (node, left, right)) = node_parser(node.as_str())?;
//         println!("{} {} {}", to_number(node), to_number(left), to_number(right));
        let node = to_number(node) as usize;
        graph[node] = [to_number(left), to_number(right)];
        match node % 26 {
            0 => {ghost_starts.push(node);},
            25 => {ghost_ends.insert(node);},
            _ => ()
        };
        if node == end {
            has_human_end = true
        }
    }
    
    let count = if has_human_end {
        let mut node = to_number("AAA") as usize;
        let mut count = 1;
        for direction in directions.chars().cycle() {
            match direction {
                'L' => { node = graph[node][0] as usize; }
                'R' => { node = graph[node][1] as usize; }
                _ => return Err(format!("Unexpected direction {}", direction).into())
            }
            
            if node == end {
                break
            }
            count += 1;
        }
        Some(count)
    } else {
        None
    };
    
    let mut graph_descriptions = Vec::new();

    for node in ghost_starts.into_iter() {
        let mut node = node;
        let mut step_counter = HashMap::new();
        let mut ghost_end = None;
        let mut count: u128 = 0;
        let mut loop_start = None;
        for (i, direction) in directions.chars().enumerate().cycle() {
            if step_counter.contains_key(&(i, node)) {
                loop_start = Some((i, node));
                break;
            }
            step_counter.insert((i, node), count);
            match direction {
                'L' => { node = graph[node][0] as usize; }
                'R' => { node = graph[node][1] as usize; }
                _ => return Err(format!("Unexpected direction {}", direction).into())
            }
            
            count += 1;
            if ghost_ends.contains(&node) {
                if let Some(_)  = ghost_end {
                    return Err(format!("Only one ghost end per ghost path supported").into());
                }
                ghost_end = Some(count);
            }
        }
        let loop_start = *step_counter.get(&loop_start.unwrap()).unwrap();
        let ghost_end = ghost_end.unwrap();
        if ghost_end < loop_start {
            return Err(format!("ghost end in preamble is not supported").into());
        }
        let loop_size = count - loop_start;
        graph_descriptions.push(loop_size);
    }
    
    println!("{:?}", graph_descriptions);
    
    let res = graph_descriptions.into_iter().reduce(|a,b| a*b/(bezout_identity(a as i128,b as i128).2 as u128)).unwrap();
    
//     let (mut loop_size, mut end) = graph_descriptions.next().unwrap();
//     while let Some((next_loop_size, next_end)) = graph_descriptions.next() {
//         println!("{} {}", loop_size, next_loop_size);
//         let (c0, c1, gcd) = bezout_identity(loop_size, next_loop_size);
//         let (c0, c1, gcd) = if end <= next_end {
//             (c0 % (next_loop_size / gcd), (-c1) % (loop_size / gcd), gcd)
//         } else {
//             ((-c0) % (next_loop_size / gcd), c1 % (loop_size / gcd), gcd)
//         };
//         
//         let diff = end.abs_diff(next_end) as i128;
//         let extra_loops = (c0 * diff / loop_size).min(c1 * diff / next_loop_size);
//         let new_loop_size = loop_size * next_loop_size / gcd;
//         println!("{} {} {}", new_loop_size, c0, extra_loops);
//         end += c0 * diff * (loop_size / gcd) - new_loop_size * extra_loops;
//         loop_size = new_loop_size;
//     }
    
    println!("{}", res);
    
    /*
    println!("{:?}", Vec::from(&graph[..10]));
    println!("{:?} {:?} {:?}", graph[0], graph[to_number("BBB") as usize], graph[to_number("CCC") as usize]);*/

    Ok([count.map(|c| c.to_string()), Some(res.to_string())])
}

#[test]
pub fn test_day8() {
    assert!(common::run_test(DAY, &run))
}
